unit global;

interface

uses Forms, Messages, Windows, Classes, Controls, Dialogs, sysutils;

var
   WaitCount : Integer;
   ResultCode : Integer;

   procedure Wait;
   procedure WaitDone;

   procedure ShellCopyFile(ParentWnd : HWND; Source : String; Dest : String);
   function IsDirectory(S : String) : Boolean;
   function FormatFileTime(FileTime : TFileTime) : String;
   function StrToFileTime(S : String) : TFileTime;
   function FileTimeToStr(Time : TFileTime) : String;
   function GetTempPath : String;
   function GetWindowsPath : String;
   function GetSystemPath : String;
   function GetSystemDrive : String;
   function FormatISODate(Date : TDateTime) : String;
   function Format1000(Number : Int64) : String;
   function FormatHuman(Number : Int64) : String;
   function GetVolumeID(DriveChar: Char): string;
   function GetSpecialPath(CSIDL: Integer) : String;
   function GetCurrentYear : Integer;
   function EndsWith(S : String; E : String) : Boolean;
   function StartsWith(S : String; E : String) : Boolean;
   function SHGetFolderPath(hwndOwner: HWND; nFolder: Integer; hToken : THANDLE; Flags : DWORD; Path : PCHAR): HResult; stdcall;
   procedure FindFiles(Path : String; Files : TStringList; Extensions : TStringList; AndExtensions : Boolean);
   function SingleInstance(Name : String) : Boolean;

type
  CancelException = class(Exception)
  end;
  
const CSIDL_LOCAL_APPDATA = $001c;        // <user name>\Local Settings\Applicaiton Data (non roaming)

implementation

uses shellapi, shlobj, activex;

const SHGFP_TYPE_CURRENT  = 0;   // current value for user, verify it exists
const SHGFP_TYPE_DEFAULT  = 1;   // default value, may not exist

function SHGetFolderPath; external 'shell32.dll' name 'SHGetFolderPathA';


procedure Wait;
begin
   Inc(WaitCount);
   if WaitCount > 0 then
   begin
      Screen.Cursor := crHourGlass;
   end
   else
   begin
      Screen.Cursor := crDefault;
   end;
end;

procedure WaitDone;
begin
   Dec(WaitCount);
   if WaitCount > 0 then
   begin
      Screen.Cursor := crHourGlass;
   end
   else
   begin
      Screen.Cursor := crDefault;
   end;
end;

procedure ShellCopyFile(ParentWnd : HWND; Source : String; Dest : String);
var
   FileOp      : TSHFILEOPSTRUCT;
   FromFile    : PChar;
   ToFile      : PChar;
begin
   if FileExists(Dest) then
   begin
      DeleteFile(Dest);
   end;

   FromFile := StrAlloc(Length(Source) + 2);
   ZeroMemory(FromFile, Length(Source) + 2);
   Move(PChar(Source)^, FromFile^, Length(Source));

   ToFile := StrAlloc(Length(Dest) + 2);
   ZeroMemory(ToFile, Length(Dest) + 2);
   Move(PChar(Dest)^, ToFile^, Length(Dest));

   ZeroMemory(@FileOp, SizeOf(FileOp));
   FileOp.Wnd := ParentWnd;
   FileOp.wFunc := FO_COPY;
   FileOp.pFrom := FromFile;
   FileOp.pTo := ToFile;
   FileOp.fFlags := FOF_NOCONFIRMATION or FOF_SIMPLEPROGRESS;
   FileOp.lpszProgressTitle := 'Creating database snapshot.  Please wait...';

   SHFileOperation(FileOp);

   if FileOp.fAnyOperationsAborted then
   begin
      if not CopyFile(PChar(Source), PChar(Dest), False) then
      begin
         ShowMessage('Fallback copy failed with error=' + SysErrorMessage(GetLastError));
      end;
   end;

   StrDispose(FromFile);
   StrDispose(ToFile);
end;

function IsDirectory(S : String) : Boolean;
var
   Attrs : Integer;
begin
   Attrs := FileGetAttr(S);
   if Attrs = -1 then
   begin
      Result := False; // if it does not exist then it can't be a directory
   end
   else
   begin
      if Attrs and faDirectory <> 0 then
      begin
         Result := True;
      end
      else
      begin
         Result := False;
      end;
   end;
end;

function FormatFileTime(FileTime : TFileTime) : String;
var
   LocalTime : TFileTime;
   SystemTime : TSystemTime;

   function FormatInt(n : Integer; width : Integer) : String;
   var
      S : String;
   begin
      S := IntToStr(n);
      while Length(S) < width do
      begin
         S := '0' + S;
      end;
      Result := S;
   end;
begin
   FileTimeToLocalFileTime(FileTime, LocalTime);
   FileTimeToSystemTime(LocalTime, SystemTime);
//   Result := Format('%4d-%2d-%2d %2d:%2d:%2d', [SystemTime.wYear, SystemTime.wMonth, SystemTime.wDay, SystemTime.wHour, SystemTime.wMinute, SystemTime.wSecond]);
   Result := FormatInt(SystemTime.wYear, 4) + '-' +
             FormatInt(SystemTime.wMonth, 2) + '-' +
             FormatInt(SystemTime.wDay, 2) + ' ' +
             FormatInt(SystemTime.wHour, 2) + ':' +
             FormatInt(SystemTime.wMinute, 2) + ':' +
             FormatInt(SystemTime.wSecond, 2);
end;

function StrToFileTime(S : String) : TFileTime;
var
   Low : String;
   High : String;
begin
   if Length(S) <> 16 then
   begin
      raise Exception.Create('Invalid FileTime String');
   end;
   High := '$' + Copy(S, 1, 8);
   Low := '$' + Copy(S, 9, 8);
   Result.dwLowDateTime := StrToInt(Low);
   Result.dwHighDateTime := StrToInt(High);
end;

function FileTimeToStr(Time : TFileTime) : String;
var
   Low : String;
   High : String;
begin
   Low := IntToHex(Time.dwLowDateTime, 8);
   High := IntToHex(Time.dwHighDateTime, 8);

   Result := High + Low;
end;

function GetTempPath : String;
var
   L : Integer;
begin
   L := 256;
   SetLength(Result, L);
   L := Windows.GetTempPath(L, PChar(Result));
   SetLength(Result, L);
end;

function GetWindowsPath : String;
var
   L : Integer;
begin
   L := MAX_PATH;
   SetLength(Result, L);
   L := Windows.GetWindowsDirectory(PChar(Result), l);
   SetLength(Result, L);
end;

function GetSystemPath : String;
var
   L : Integer;
begin
   L := MAX_PATH;
   SetLength(Result, L);
   L := Windows.GetSystemDirectory(PChar(Result), l);
   SetLength(Result, L);
end;

function GetSystemDrive : String;
var
   S : String;
begin
   S := GetSystemPath;
   Result := Copy(S, 1, 3);
end;

function FormatISODate(Date : TDateTime) : String;
begin
   Result := FormatDateTime('yyyy-mm-dd', Date);
end;

function Format1000(Number : Int64) : String;
var
   S : String;
begin
   // Like IntToStr but returns numbers like 123,456,789
   S := IntToStr(Number);
   while Length(S) > 3 do
   begin
      Result := ',' + Copy(S, Length(S) - 2, 3) + Result;
      Delete(S, Length(S) - 2, 3);
   end;
   Result := S + Result;
end;

function FormatHuman(Number : Int64) : String;
var
   Number2 : single;
begin
   // Like IntToStr but returns numbers like 100kb, 2Mb etc.
   if Number < 1024 then
   begin
      Result := IntToStr(Number) + ' bytes';
   end
   else if Number < 1048576 then
   begin
      Number := Number div 1024;
      Result := IntToStr(Number) + ' kb';
   end
   else
   begin
      Number2 := Number / 1048576;
      Result := Format('%.1f', [Number2]) + ' Mb';
   end;
end;

function GetVolumeID(DriveChar: Char): string;
var
  OldErrorMode: Integer;
  NotUsed, VolFlags: DWORD;
  Buf: array [0..MAX_PATH] of Char;
begin
   if DriveChar = 'A' then
   begin
      Result := '';
   end
   else
   begin
     OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
     try
       Buf[0] := #$00;
       if GetVolumeInformation(PChar(DriveChar + ':\'), Buf, DWORD(sizeof(Buf)),
         nil, NotUsed, VolFlags, nil, 0) then
         SetString(Result, Buf, StrLen(Buf))
       else Result := '';
       if DriveChar < 'a' then
         Result := AnsiUpperCaseFileName(Result)
       else
         Result := AnsiLowerCaseFileName(Result);
     finally
       SetErrorMode(OldErrorMode);
     end;
   end;
end;

function GetSpecialPath(CSIDL: Integer) : String;
var
   Path : String;
   idl : PItemIDList;
   ppMalloc : IMalloc;
begin
   Result := '';

   SetLength(Path, MAX_PATH);
   if SHGetSpecialFolderLocation(0, CSIDL, idl) = NOERROR then
   begin
//   if SHGetSpecialFolderPath(0, PChar(Path), CSIDL{_PERSONAL}, False)
//   if SHGetFolderPath(0, CSIDL, 0, SHGFP_TYPE_CURRENT, PChar(Path)) = S_OK then
      if SHGetPathFromIDList(idl, PChar(Path)) then
      begin
         SetLength(Path, strlen(PChar(Path)));
         Path := Path + '\';
//          MessageDlg(Path, mtInformation, [mbOK], 0);
         Result := Path;
      end
      else
      begin
//      Error := GetLastError;
//      MessageDlg(SysErrorMessage(Error), mtError, [mbOK], 0);
      end;

      // we should free the idl
      if SHGetMalloc(ppMalloc) = NOERROR then
      begin
         ppMalloc.Free(idl);
      end;
   end;
end;

function GetCurrentYear : Integer;
var
   Year  : WORD;
   Month : WORD;
   Day   : WORD;
begin
   DecodeDate(Now, Year, Month, Day);
   Result := Year;
end;

// returns true if S ends with E
function EndsWith(S : String; E : String) : Boolean;
var
   ELen : Integer;
begin
   ELen := Length(E);
   if LowerCase(Copy(S, Length(S) - ELen + 1, ELen)) = LowerCase(E) then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

// returns true if S starts with E
function StartsWith(S : String; E : String) : Boolean;
begin
   if Pos(LowerCase(E), LowerCase(S)) = 1 then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

// extensions is a list of file extensions, including the dot.  The object should be nil
// if the object pointer is 1 then the start of the file name is matched
procedure FindFiles(Path : String; Files : TStringList; Extensions : TStringList; AndExtensions : Boolean);
   procedure Search(Path : String);
   var
      FindFileData : TWIN32FindData;
      hFind : THandle;
      Name : String;
      SearchPattern : String;
      i : Integer;
      Matches : Integer;
   begin
      if IsDirectory(Path) then
      begin
         if not EndsWith(Path, '\') then
         begin
            Path := Path + '\';
         end;
         SearchPattern := Path + '*.*';
         hFind := FindFirstFile(PChar(SearchPattern), FindFileData);
         while hFind <> INVALID_HANDLE_VALUE do
         begin
            Name := FindFileData.cFileName;
            if (Name = '.') or (Name = '..') then
            begin
               // skip...
            end
            else if (FindFileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = FILE_ATTRIBUTE_DIRECTORY then
            begin
               Search(Path + Name);
            end
            else
            begin
               Matches := 0;

               for i := 0 to Extensions.Count - 1 do
               begin
                  if Integer(Extensions.Objects[i]) = 1 then
                  begin
                     if StartsWith(Name, Extensions[i]) then
                     begin
                        Matches := Matches + 1;
                     end;
                  end
                  else
                  begin
                     if EndsWith(Name, Extensions[i]) then
                     begin
                        Matches := Matches + 1;
                     end;
                  end;
               end;

               if AndExtensions then
               begin
                  if Matches = Extensions.Count then
                  begin
                     Files.Add(Path + Name);
                  end;
               end
               else // and
               begin
                  if Matches > 0 then
                  begin
                     Files.Add(Path + Name);
                  end;
               end;
            end;

            if not FindNextFile(hFind, FindFileData) then
            begin
               Windows.FindClose(hFind);
               hFind := INVALID_HANDLE_VALUE;
            end;
         end;
      end;
   end;
begin
   Search(Path);
end;

function SingleInstance(Name : String) : Boolean;
var
   LockHandle : DWORD;
   Path : String;
begin
   Result := False;
   Path := GetTempPath + '\' + Name + '.lock';
   LockHandle := CreateFile(Pchar(Path), GENERIC_READ or GENERIC_WRITE, 0, nil ,OPEN_ALWAYS, 0, 0);
   if LockHandle = INVALID_HANDLE_VALUE then
   begin
      Result := True;
      MessageDlg(Name + ' is already running.  You can only run one instance at a time', mtError, [mbOK], 0);
      Application.Terminate;
      raise Exception.Create('Application Terminating');
   end;
end;

initialization
   WaitCount := 0;
   ResultCode := 0;

finalization
//   Global.Terminate;


end.
